
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>


enum {
//	_xinfreq = 20_000_000,
	_xtlfreq = 20_000_000,
	_clkfreq = 4_000_000,
	DOWNLOAD_BAUD = 230_400,
	DEBUG_BAUD = DOWNLOAD_BAUD,
/*
	spi_cs = 61,
	spi_ck = 60,
	spi_di = 59,
	spi_do = 58,
*/
	spi_ck = 56,
	spi_di = 57,
};



uint32_t  crc7( uint8_t * buf, int len )
{
	uint32_t  val, crc = 0;

	__asm {
// Reference code courtesy of Ariba
cr7lp
		rdbyte	val, buf
		add	buf, #1
		shl	val, #24
		setq	val
		crcnib	crc, #$90>>1
		crcnib	crc, #$90>>1
		djnz	len, #cr7lp

		rev	crc
		shr	crc, #24
		or	crc, #1
	}
	return crc;
}



uint32_t  crc16( uint32_t * buf, int len )
{
	uint32_t  val, crc = 0,
	polynomial = 0x8408;  // CRCNIB shifts right, therefore ordered from left: x0 + x5 + x12

	__asm volatile {		// "volatile" prevents optimisation
// Reference code is courtesy of Ariba
		rdfast	#0, buf
		mov	ptra, #0

		rep	@.rend, len
		rflong	val
		setq	val
		crcnib	crc, polynomial
		crcnib	crc, polynomial
		crcnib	crc, polynomial
		crcnib	crc, polynomial
		crcnib	crc, polynomial
		crcnib	crc, polynomial
		crcnib	crc, polynomial
		crcnib	crc, polynomial
.rend
		rev	crc
		shr	crc, #16
	}
	return crc;
}




static uint32_t  xmit_blk( 
	const uint8_t * buf,	// Data to be sent
	size_t bc,		// Number of longwords to send
	uint32_t ratio )	// sysclocks per SPI clock
{
	const int  PIN_CLK = 56;
	const int  PIN_DI = 52 | 3<<6;
	const uint8_t  phase[] = { 0,0,0,0, 1,3,5,0, 0,1,2,3, 4,5,6,8, 8,9,10,11 };
	uint32_t  crc = phase[ratio - 1];
	uint32_t  polynomial = X_RFBYTE_4P_1DAC4 | X_PINS_ON | X_ALT_ON | ((PIN_DI&0x3c)<<17);

	_pinf( PIN_CLK );
	_wrpin( PIN_CLK, P_PULSE | P_OE | P_INVERT_OUTPUT );
	_wxpin( PIN_CLK, ratio | (ratio>>1)<<16 );

	_wrpin( PIN_DI, P_SYNC_IO );
	_pinl( PIN_DI );

	ratio = 0x8000_0000UL / ratio;

	__asm volatile {            // "volatile" prevents optimisation
//		qdiv	##$8000_0000, ratio

//		fltl	PIN_CLK
//		wrpin	##P_PULSE | P_OE | P_INVERT_OUTPUT, PIN_CLK    // CPOL=1
//		mov	pb, ratio
//		shr	pb, #1
//		setword	ratio, pb, #1 
//		wxpin	ratio, PIN_CLK

//		wrpin	##P_SYNC_IO, PIN_DI
//		drvl	PIN_DI

		shl	bc, #3    // nibble count
		or	polynomial, bc
		rdfast	#0, buf

//		getqx	ratio
		setxfrq	ratio
//timing critical code
		dirh	PIN_CLK
		waitx	crc
		xinit	polynomial, #0
		wypin	bc, PIN_CLK    // 22 longwords
//end of critical code

//		shr	bc, #3    // longword count
//		mov	bc, #4    // loop count
		mov	crc, #0
		mov	polynomial, ##$8408    // from the left: x0 + x5 + x12
//		mov	ratio, ##64<<9    // 
//		mov	pb, ##7<<3 | 3<<12    // %000_011_000_000_111_000, post-incrementing and modulo 64
		setq2	#127
		rdlong	0, buf    // copy block (512 bytes) to lutRAM 0..127
		mov	ptrb, #0
.crcloop
// The following block copy assumes unused cogRAM at addresses 64..127
//		setq	#63
//		rdlong	64, buf    // copy half block (256 bytes) to cogRAM 64..127
//		add	buf, #256

		rep	@.rend, #128   // longword count
//		alti	ratio, #$38    // %000_111_000, post-incrementing register number for SETQ
//		alti	ratio, pb
//		setq	0-0    // put data in Q register for CRCNIB
		rdlut	ratio, ptrb++
		setq	ratio
		crcnib	crc, polynomial
		crcnib	crc, polynomial
		crcnib	crc, polynomial
		crcnib	crc, polynomial
		crcnib	crc, polynomial
		crcnib	crc, polynomial
		crcnib	crc, polynomial
		crcnib	crc, polynomial
.rend
//		djnz	bc, #.crcloop

//		waitx	#1
//		nop
//		nop
//		nop
//		nop
//		nop
		rev	crc
		shr	crc, #16
		waitxfi
	}
	return crc;
}

/*
static uint32_t  xmit_blk( 
	const uint8_t * buf,	// Data to be sent
	size_t bc,		// Number of longwords to send
	unsigned ratio )	// sysclocks per SPI clock
{
	uint32_t  crc = 0;

	__asm volatile {            // "volatile" prevents optimisation
		setxfrq	##$1000_0000
		rdfast	#0, #0
		xinit	##X_RFBYTE_4P_1DAC4 | X_PINS_ON | X_ALT_ON | 52<<17 | 200, #0

		setq2	#127
		rdlong	0, buf    // copy block (512 bytes) to lutRAM

		waitx	#10
		rev	crc
		shr	crc, #16
		waitxfi
	}
	return crc;
}
*/

void  main( void )
{
	uint8_t  buff[512] = {0x40,0x00,0x00,0x00,0x00};
	unsigned  i, ticks;

	_waitms( 200 );
	printf( "   clkfreq = %d   clkmode = 0x%x\n", _clockfreq(), _clockmode() );
	printf( "crc7 of buff[] = %x\n", crc7( buff, 5 ) );

	for( i = 0; i < sizeof(buff); i++ )
	{
		buff[i] = 255;
	}

	i = 8;
	printf( "data rate = sysclock/%d\n", i );
	_waitms( 200 );

	ticks = _cnt();
	i = xmit_blk( buff, sizeof(buff)/4, i );
	ticks = _cnt() - ticks;

	printf( "crc16 of buff[] = %x, completed in %d ticks\n", i, ticks );

}
